# Kaskade-S/T 
# v 1.1: 2011-05-04
# - This program needs Python 3.1 or higher to execute
# 
# Encryption and decryption with a cascade of one substitution and one transposition:
# - Each part of the cascade uses its own permutation of numbers as its key.
#   The overall "key" is the sequence of two permutations of the same length.
#   The first perm is used by the substitution, the second perm by the transposition.
#   A full cascade can be applied several times (according to the specified rounds).
# - Keys can be generated using two random or two identical permutations.
#   Keys can be printed to a file in readable form as blank separated integers,
#   and a integer sequence in a file can be transformed to a key file.
# - The length of a permutation is also the length of the alphabet. Default is 256.
#   If you specify a shorter length and a given offset (using -f) for the alphabet,
#   then the input files (plaintext or ciphertext) may ONLY contain characters
#   of the specified alphabet. E.g. if you use only the 26 uppercase letters:
#      python Kaskade.py  -g Random-Different  -l 26
#      python Kaskade.py  -e plaintext ciphertext  Key  -l 26  -f A
#      python Kaskade.py  -d ciphertext  plaintext-dec  Key  -l 26  -f A
#
# Remark:
# - Here it doesn't matter which encryption method is applied first.
# - All input comes from files (despite the options).


### Imports
### =======
import random
import os
import re

### Default parameter settings
### (use extra global control variables if value depends on more than one command
###  line option, as the order of the options is variable)
### ==========================
T_firstAlphChar = False	 # True if you specify the smallest character of the alphabet.
			 # Use case e.g.: Alphabet is capital letters only.
                         #   (True: when -f and -l).
first=0		         # Specify the first letter of a shortend alphabet
                         #   (must be 0 as default).
lop=256                  # Length of one permutation (default=256;
                         #        can be made smaller via option -l)
                         # The whole key exists of two permutations.
                         # Later on, data blocks are handled in the size of lop.
roundsCascades=1         # Specify the number of rounds (must be 1 as default).
                         # Must be set, because the default is used, even
			 # if the option opts.rounds (-r ) is not set.
T_ReadKeyFromFile=False  # True: when (-e or -d) and 3 parameters are given
T_RandomPermKey=False    # True: when -g and key random (instead of identical perm)
T_SamePermKeys=False     # True: when -g and key parts are different (instead of two same perms)

Name_f_keyout = "Key"    # The two permutations created are written into a key file (used only when -g).
Name_f_KeySeparatedNumbers = "KeyAsNumbers.txt"  # The key (just generated or read) is written into a key file as separated single numbers (used only when -p).




### Used functions
### ==============

# Create a permutation of length k
# - Input:  length k, bool RandomPermKey (True = random perm; False = identical perm)
# - Output: perm  (one perm is used as one half of the key)
def CreatePerm(k, RandomPermKey):
  if not RandomPermKey:
    return [i%k for i in range(k)]  # List comprehension
  else:
    return random.sample(range(k),k)


# Invert a given permutation
# - Input:  length k of the permutation, given permutation
# - Output: inverted permutation
def InversePerm(k, perm_in):
  perm_out=perm_in[:]
  for i in range(k):   
     perm_out[perm_in[i]]=i 
  return perm_out


# Verify that all k numbers appear in the specified permutation
# - Input:  length k of the permutation, permutation to validate
# - Output: True if perm is valid; False if perm is invalid
def ValidatePerm(k, perm_in):
  count = [0] * k
  
  for i in range(len(perm_in)):
    entry = perm_in[i]
    if (entry < 0 or entry >= k):
      print("Validation error: entry", entry, "out of range (0 ..", k - 1, ") at offset", i)
      return False
    count[entry] += 1
    if count[entry] > 1:
      print("Validation error: duplicate value", entry, "at offset", i)
      return False

  for i in range(k):
    if (count[i] < 1):
      print("Validation error: value", count[i], "is missing in permuation")
      return False

  return True


# Print key (store the two permutations as separated numbers to a file)
# - Input:  length k, permutations of length k permS and permT as the two parts of the key
# - Output: only writing to file
def StoreKeyAsNumbersToFile(permS, permT, k):
  stors = list(range(k))

  with open(Name_f_KeySeparatedNumbers, "w") as f_KeySeparatedNumbers:

    stors[0:k]=[int(permS[i]) for i in range(k)]
    f_KeySeparatedNumbers.writelines(list("%3s " % item for item in stors))

    f_KeySeparatedNumbers.writelines("\n")

    stors[0:k]=[int(permT[i]) for i in range(k)]
    f_KeySeparatedNumbers.writelines(list("%3s " % item for item in stors))


# Perform substitution encryption or substitution decryption
# - Cipher applied to one data block of length k,
# - Input:  block length k, perm as key, data block
#           (for encryption the function gets the stored substitution perm;
#            for decryption it gets the inverted perm) 
# - Output: substituted data block
def EncDecSubst(k, perm, DataIn, DataOut):
  DataOut[0:k]=[perm[DataIn[i]] for i in range(k)]  # List comprehension


# Perform transposition encryption or transposition decryption
# - Cipher applied to one data block of length k,
# - Input:  block length k, perm as key, data block 
#           (for encryption the function gets the stored transposition perm;
#            for decryption it gets the inverted perm) 
# - Output: transpositioned data block
def EncDecTranspos(k, perm, DataIn, DataOut):
  for i in range(k):
    DataOut[perm[i]]=DataIn[i] 




### Handle Options (command line parameters)
### ========================================

## Get options
## ===========

from optparse import OptionParser
pars = OptionParser(version='%prog version 1.1, May 2011 (open source, written for the MTC3 contest)')

pars.add_option("-e", "--encrypt", help='Encrypts file_in, builds file_out, uses keyfile_in as key.',\
                dest='T_EncryptFileArg', default=False, action='store',\
                metavar='<file_in> <file_out> <keyfile_in>', nargs=3, type='string')

pars.add_option("-d", "--decrypt", help='Decrypts file_in, builds file_out, uses keyfile_in as key.',\
                dest='T_DecryptFileArg', default=False, action='store',\
                metavar='<file_in> <file_out> <keyfile_in>',  nargs=3, type='string')

pars.add_option("-l", "--permlen", help='Defines length of one permutation. Default is 256. Can be made shorter. A key consists of two permutations.', type='int') 

pars.add_option("-f", "--first", help='Defines the first character of an alphabet shorter than 256. The used alphabet then starts at the given character.', type='str') 

pars.add_option("-r", "--rounds", help='Defines the number of rounds how often to perform the whole cascade of one substitution and one transposition. Default is 1.', type='int') 

helptext4g='Generates a key and stores it in the file "' + Name_f_keyout +'". There are 3 possible arguments to specify the generated key: Random-Different, Random-Same, Ident.'
pars.add_option("-g", "--genkey", help=helptext4g,\
                dest='T_genkey', default=False, action='store',\
                metavar='<RandomOrNot-DiffOrSame>',  nargs=1, type='string')

pars.add_option("-p", "--printkey",\
	   help='Stores the key as separated numbers in the file "'+ Name_f_KeySeparatedNumbers +'", each permutation in one line.',\
           dest='T_printkey', default=False, action='store_true')

pars.add_option("-a", "--readintegers",\
                  help='Reads a file with separated integer numbers and builds a key file.',\
                  dest='T_readintegers', default=False, action='store', metavar='<number_keyfile>', nargs=1, type='string')

pars.add_option("-v", "--verbose", help='Print commandline options, cwd, and some intermediate results (makes sense only together with other options).',\
                dest='T_verbose', default=False, action='store_true')
                # If True: Show some intermediate results too (option -v).

(opts, args) = pars.parse_args()



## Interpret/check options, exit if wrong, set according global variables
## - here only perform pure output like --verbose or --help.
## ======================================================================

# Verbose: -v
if opts.T_verbose:
  print("CWD =", os.getcwd())
  print( "Opts:", opts )

# Specify the first letter of a shortend alphabet: -f
if opts.first:
  if opts.permlen:
    first = ord(opts.first[0])
    T_firstAlphChar = True
  else:
    print("Error: Option -l is missing.")
    exit()

# Encryption and decryption: -e and -d
if opts.T_EncryptFileArg or opts.T_DecryptFileArg:
  T_ReadKeyFromFile=True
  
  if opts.T_EncryptFileArg:
    (Name_f_in, Name_f_out, Name_f_permin) = opts.T_EncryptFileArg[0:3]  # Tuple unpacking using slicing of the list 

  if opts.T_DecryptFileArg:
    (Name_f_in, Name_f_out, Name_f_permin) = opts.T_DecryptFileArg[0:3]  # Tuple unpacking using slicing of the list

  if not os.path.exists(Name_f_in):
    print("Error: File \"", Name_f_in, "\" does not exist.")
    exit()

  if not os.path.exists(Name_f_permin):
    print("Error: File \"", Name_f_permin, "\" does not exist.")
    exit()


# Set length of one permutation: -l
if opts.permlen:
  lop = opts.permlen
  if lop < 0 or lop > 256:
    print("Error with -l: Given perm length must not be negative or bigger than 256.")
    exit()
  if lop!=256 and not(opts.first) and not(opts.T_genkey):
    print("Error: Setting the perm length requires to specify option -f too (when doing enc or dec), as the length of the permutation defines both half of the key length as well as the maximal length of the used alphabet.")
    exit() 


# Set number of rounds: -r
if opts.rounds:
  roundsCascades = opts.rounds
  if roundsCascades < 0:
    print("Error with -r: Given number of rounds must be a positive integer.")
    exit()


# Generate key: -g
if opts.T_genkey:
  dict = {"Random-Different" : (True,  False),  # Dictionary instead of if-else
          "Random-Same"      : (True,  True ),  # If random, the 2 perms could be diff.
          "Ident"            : (False, True) }  # Both identical perm are the same.
  if opts.T_genkey not in dict:
    print("Error with -g: Argument \"", opts.T_genkey, "\" is not allowed.")
    exit()
  else:
    (T_RandomPermKey, T_SamePermKeys) = dict[opts.T_genkey]


# Print key: -p
# - Nothing to do by itself. 
# - If opts.T_printkey==True and a key is generated or read (either as Key file or from separated
#   numbers, it will be additionally stored as separated numbers into the file named in the
#   variable Name_f_KeySeparatedNumbers (within the function StoreKeyAsNumbersToFile()).



# Raise exception: options is neither of the following: -e, -d; -g, -a
if not (opts.T_EncryptFileArg or opts.T_DecryptFileArg or opts.T_genkey or opts.T_readintegers):
  print("Error: Please enter a valid option: -e, -d, -g, -a.")
  exit()


## Interpret options, set according global variables, process options
## (if more than one option is involved)

if opts.T_verbose and T_firstAlphChar:
  print(first)



## Deal with the "standalone" actions (no encryption or decryption)
## ================================================================

## Generate a key, store it in a file and exit (only when -g)
## - create the key (it consists of two permutations which have the same length lop)
## - each permutation is either a random permutation or an identical permutation
## - if T_SamePermKeys==True both newly created perms are the same.
if opts.T_genkey:  # comparing opts.T_genkey==True is wrong
  permS=CreatePerm(lop, T_RandomPermKey)
  if T_SamePermKeys==False:
    permT=CreatePerm(lop, T_RandomPermKey)
  else:
    permT=permS

  with open(Name_f_keyout, "wb") as f_keyout:  # with open saves a f_keyout.close()
    f_keyout.write(bytearray(permS))
    f_keyout.write(bytearray(permT))

  if T_RandomPermKey==True and T_SamePermKeys==False:
    print( "Two different random perms, each of length", lop, "written as key to file", Name_f_keyout, ".")
  elif T_RandomPermKey==True and T_SamePermKeys==True:
    print( "Two same random perms, each of length", lop, "written as key to file", Name_f_keyout, ".")
  elif T_RandomPermKey==False and T_SamePermKeys==True:
    print( "Two same identical perms, each of length", lop, "written as key to file", Name_f_keyout, ".")

  if opts.T_printkey == True:
    StoreKeyAsNumbersToFile(permS, permT, lop)

  exit()



## Read two permutations from a file with separated decimal ascii integers,
## store it as key in a file and exit (only when -a).
if opts.T_readintegers:
  Name_k_in = opts.T_readintegers
  if not os.path.exists(Name_k_in):
    print("Error: File \"", Name_k_in, "\" does not exist.")
    exit()

  data = []
  
  # regex parser that splits on decimal number boundaries (in ascii)
  # - all non decimal digits are considered as separators (including blanks, comma, etc)
  p = re.compile(r'\D+')

  keyfile  = open(Name_k_in, "rU")
  for line in keyfile:
    row = p.split(line)

    for item in row:
      if (item != ""):
        data.append(int(item))

  if (len(data) % 2 != 0):
    print("Data file contains odd number of elements. It should contain an even number of integers.")
    exit()

  k = int(len(data) / 2);

  # split input data evenly into two key parts
  permS = data[:k]
  permT = data[k:]

  if opts.T_verbose:
    print("permS:     ", permS)
    print("permT:     ", permT)

  if (not ValidatePerm(k, permS)):
    print("Error: First part of supposed key is invalid.")
    exit()

  if (not ValidatePerm(k, permT)):
    print("Error: Second part of supposed key is invalid.")
    exit()

  print('Integers of two k-permutations were successfully read from file', Name_k_in, '. k=', k)

  with open(Name_f_keyout, "wb") as f_keyout:  # with open saves a f_keyout.close()
    f_keyout.write(bytearray(permS))
    f_keyout.write(bytearray(permT))

  if opts.T_printkey == True:
    StoreKeyAsNumbersToFile(permS, permT, k)

  print("Key was written to file", Name_f_keyout)
  exit()




### Main Program
### - performs encryption or decryption (only when -e or -d).
### - reads the key and the input (plaintext or ciphertext) each from a file.
### ============

# Create physical storage for a list to be used for generating further storage (via slicing)
stor = list(range(lop))

# Allocate memory for the two permutations
permS=stor[:]
permT=stor[:]

# Read key from file
if T_ReadKeyFromFile==True:
  with open(Name_f_permin, "rb") as f_permin:  # with open saves a f_permint.close()
    permS=f_permin.read(lop)
    permT=f_permin.read(lop)
  permS= list(permS) 
  permT= list(permT) 

  if len(permS+permT)!= 2*lop:
    print("Info: Length of key file = " , len(permS+permT), ".")
    if lop != 256:
      print("Error: Parameter from option -l (", lop, ") is wrong. It should be half of key length.")
    else:
      print("Error: Length of key file does not match to the expected length of 2*256 bytes.")
    exit()

  if opts.T_printkey == True:
    StoreKeyAsNumbersToFile(permS, permT, lop)

  if opts.T_DecryptFileArg:
    permS=InversePerm(lop, permS)
    permT=InversePerm(lop, permT)

  if opts.T_verbose:
    print("permS:     ", permS)
    print("permT:     ", permT)

# Allocate memory for one data blocks
DataOut=stor[:]

# Open input and output file
f_in  = open(Name_f_in, "rb")
f_out = open(Name_f_out, "wb")

# Length of the output file (changed during the main loop)
kkoF = 0


## Main loop: Process each block with length lop until EOF
##            - each data block is handled independantly.
## =======================================================
while True:

  roundsBlock = roundsCascades   # The number of rounds is the same for each block.
	
  # Read from data input file
  DataIn = bytearray(f_in.read(lop)) # bytearray (read/write) rather than bytes (read-only).

  lenDataIn = len(DataIn)
  kkoF += lenDataIn
  if opts.T_verbose:
    print("Read Data=", DataIn, "len=", lenDataIn)

  # Block length == 0: This happens at the end when input length is a multiple of lop
  if lenDataIn == 0:
    break


  # Padding if: 0 < block length of read data < lop
  # Padded value must be part of the used alphabet (first, ..., first+lop-1)
  #   ( if not: problems with index in EncDecSubst(), DataOut[i]=perm[DataIn[i]] )
  # Padding: Default: with X=88 (only if X is part of the alphabet)
  #          Else   : with the first character of the alphabet.
  if 0 < lenDataIn < lop:
    ValueCharX = ord('X')
    if first <= ValueCharX < lop+first:
      DataIn=DataIn + (lop-lenDataIn)*b"X"
    else:
	    # DataIn=DataIn + (lop-lenDataIn)*b'\x00'  # Less flexible alternative: Fill up with zeros
      OneString = (lop-lenDataIn)*chr(first)
      DataIn=DataIn + bytearray(OneString.encode())
    if opts.T_verbose:
      print("Extended Data=", DataIn, ", len=", len(DataIn))
    kkoF += lop-lenDataIn


  # Now: Length of given block length == lop 
  # Within given input: Map alphabet down to range 0 to lop-1
  # (only necessary, if option -l is used; default uses complete range of a byte)
  if T_firstAlphChar:
    iTemp = 10
    for i in range(lop):
      iTemp=DataIn[i]-first  # Don't assign directly to DataIn[i] (as if input contains chars < first,
                             # then negative values are calculated, causing exception when assigned to bytes)
      # print("iTemp=", iTemp)   # for Debugging only
      if iTemp >= lop:
        print("Error: Inputdata out of range (bigger than lop=", lop, "): ", iTemp)
        exit()      
      if iTemp < 0:
        print("Error: Inputdata out of range (negative): first = ", first, ", iTemp = ", iTemp)
        if DataIn[i] == ord('\r') or DataIn[i] == ord('\n'):
          print("Error: Newline detected. This is not part of the alphabet. File may only contain characters which are in the alphabet!")
        exit()
      DataIn[i]=iTemp



  # Perform encryption or decryption for this single block (default: one round)
  while (True):
    roundsBlock = roundsBlock - 1

    EncDecSubst(lop, permS, DataIn, DataOut)
    if opts.T_verbose:
      print("SDaten:     ", DataOut)

    DataIn = DataOut[:]
    EncDecTranspos(lop, permT, DataIn, DataOut)
    if opts.T_verbose:
      print("TDaten:     ", DataOut)

    if roundsBlock <= 0:
      break
    else:
      DataIn = DataOut[:]



  # Created output: Map alphabet back to range first to first+lop-1
  # (only necessary, if option -l is used; default uses complete range of a byte)
  if T_firstAlphChar:
    DataOut[0:lop] = [x+first for x in DataOut] # more elegant than for loop

  f_out.write(bytearray(DataOut)) 


  # Leave the while loop
  if 0 < lenDataIn < lop:
    break



# Clean up
f_in.close()
f_out.close()

# Notify user
if opts.T_EncryptFileArg:
  print("Ciphertext of length", kkoF, " written to ", Name_f_out, ".")
if opts.T_DecryptFileArg:
  print("Decrypted ciphertext of length", kkoF, " written to ", Name_f_out, ".")

